#include "api-driver/structs.h"
#include "api-driver/proxy.h"
#include "common/nlohmann/json.hpp"
#include <iomanip>
#include <sys/sysinfo.h>
#include <boost/property_tree/ptree.hpp>
#include "terminal_api_driver.h"


#define TIME_NOW() std::chrono::time_point_cast<std::chrono::milliseconds>(std::chrono::system_clock::now()).time_since_epoch().count()

typedef boost::property_tree::ptree::path_type json_path;

static constexpr const char* DEFAULT_QOS_CLASSES = R"({"rt1":[],"rt2":[],"rt3":[],"cd":[]})";
// пороговое значение сна

double translateCoordinates(uint8_t deg, uint8_t min) {
    return static_cast<double>(deg) + static_cast<double>(min) / 60;
}

// static int calculateSubnetMask(const std::string& subnet_mask) {
//     int mask = 0;
//     std::istringstream iss(subnet_mask);
//     std::string octet;
//     while (std::getline(iss, octet, '.')) {
//         int octet_value = std::stoi(octet);
//         for (int i = 7; i >= 0; i--) {
//             if (octet_value & (1 << i)) {
//                 mask++;
//             }
//         }
//     }
//     return mask;
// }

/**
 * Преобразует строку вида `1.2.3.4/24` в пару строк вида `1.2.3.4` `255.255.255.0`
 */
// std::pair<std::string, std::string> splitIpAndMask(const std::string& input) {
//     auto pos = input.find('/');
//     if (pos == std::string::npos) {
//         // Обработка ошибки: нет символа '/'
//         throw std::runtime_error("address not contains mask");
//     }
//     std::string ip = input.substr(0, pos);
//     const unsigned int mask_int = std::stoul(input.substr(pos + 1));
//
//     if (mask_int > 32) {
//         throw std::runtime_error("invalid mask");
//     }
//
//     std::string mask_binary = std::string(mask_int, '1') + std::string(32 - mask_int, '0');
//     std::string mask_str;
//
//     for (unsigned int i = 0; i < 4; ++i) {
//         std::string octet = mask_binary.substr(i * 8u, 8);
//         int octet_value = std::stoi(octet, nullptr, 2);
//         mask_str += std::to_string(octet_value) + (i < 3 ? "." : "");
//     }
//
//     return std::make_pair(ip, mask_str);
// }
static inline void rtrim(std::string &s) {
    s.erase(std::find_if(s.rbegin(), s.rend(), [](unsigned char ch) {
        return !std::isspace(ch);
    }).base(), s.end());
}
static inline const char* boolAsStr(bool value) {
    return value ? "true" : "false";
}

std::ostream& operator<<(std::ostream& out, CP_Result result) {
    switch (result) {
        case OK: out << "OK"; break;
        case TIMEOUT: out << "TIMEOUT"; break;
        case ERROR: out << "ERROR"; break;
        case ABORT: out << "ABORT"; break;
        case BUSY: out << "BUSY"; break;
        default:
            out << static_cast<int>(result);
    }
    return out;
}

std::string makeTimepointFromMillis(int64_t unix_time_ms) {
    // Преобразуем миллисекунды в микросекунды для std::chrono
    auto time_point = std::chrono::time_point<std::chrono::system_clock,
        std::chrono::microseconds>(std::chrono::microseconds(unix_time_ms * 1000));

    auto tp = std::chrono::system_clock::to_time_t(time_point);
    tm* t = std::localtime(&tp);

    std::stringstream ss;
    ss << std::put_time(t, "%Y-%m-%d %H:%M:%S");
    auto ms = (unix_time_ms % 1000);
    ss << '.' << std::setw(3) << std::setfill('0') << ms;
    return ss.str();
}


#ifdef API_OBJECT_DEBUG_METRICS_ENABLE
api_driver::obj::StatisticsLogger::StatisticsLogger(): timeStart(TIME_NOW()) {}

std::string api_driver::obj::StatisticsLogger::getSettings() {
    std::lock_guard _lock(mutex);
    std::stringstream res;
    res << "{\"en\":" << boolAsStr(this->logEn);
    res << ",\"logPeriodMs\":" << logPeriodMs;
    res << ",\"maxAgeMs\":" << maxAgeMs;
    res << '}';
    return res.str();
}

void api_driver::obj::StatisticsLogger::setSettings(boost::property_tree::ptree &pt) {
    const bool newEn = pt.get<bool>("en");
    const int newInterval = pt.get<int>("logPeriodMs");
    const int newMaxAgeMs = pt.get<int>("maxAgeMs");

    std::lock_guard _lock(this->mutex);
    this->logPeriodMs = newInterval;
    this->maxAgeMs = newMaxAgeMs;

    if (newEn != this->logEn) {
        if (newEn) {
            this->logFile.open(LOG_FILENAME, std::ios::out);
            if (this->logFile.is_open()) {
                const auto* header = "timestamp\tcnt ok\tcnt bad\tfine freq dem\tcrs freq dem\tcrs freq compensator\tcrs time est\tfine time est\tmax level corr\tcurrent delay\tSNR\tcurrent modcod\tfine freq compensator\tind freq grb\tind freq tochn\tind filt adapt\tfilter corr cinc\tcorr cnt\tRSS\tcor erl\tcor lat\tgc gain\tpower pl rx\n";
                this->logFile.write(header, static_cast<std::streamsize>(strlen(header)));
                this->logEn = true;
                this->timeStart = TIME_NOW();
            }
        } else {
            if (this->logFile.is_open()) {
                this->logFile.close();
            }
            this->logEn = false;
        }
    }
}

void api_driver::obj::StatisticsLogger::updateCallback(proxy::CpProxy &cp) {

}

void api_driver::obj::StatisticsLogger::putItem(const debug_metrics &item) {
    std::lock_guard _lock(this->mutex);
    if (!logEn) return;
    if (this->logFile.is_open()) {
        std::stringstream res;
        res << makeTimepointFromMillis(TIME_NOW()) << '\t';
        res << item.cnt_ok << '\t';
        res << item.cnt_bad << '\t';
        res << item.fine_freq_dem << '\t';
        res << item.crs_freq_dem << '\t';
        res << item.crs_freq_compensator << '\t';
        res << item.crs_time_est << '\t';
        res << item.fine_time_est << '\t';
        res << item.max_level_corr << '\t';
        res << item.current_delay << '\t';
        res << item.SNR << '\t';
        res << item.current_modcod << '\t';
        res << item.fine_freq_compensator << '\t';
        res << item.ind_freq_grb << '\t';
        res << item.ind_freq_tochn << '\t';
        res << item.ind_filt_adapt << '\t';
        res << item.filter_corr_cinc << '\t';
        res << item.corr_cnt << '\t';
        res << item.RSS << '\t';
        res << item.cor_erl << '\t';
        res << item.cor_lat << '\t';
        res << item.gc_gain << '\t';
        res << item.power_pl_rx << '\n';

        const auto out = res.str();
        this->logFile.write(out.c_str(), static_cast<std::streamsize>(out.length()));
        this->logFile.flush();
    }
}

api_driver::obj::StatisticsLogger::~StatisticsLogger() = default;
#endif

#ifdef API_OBJECT_NETWORK_SETTINGS_ENABLE
api_driver::obj::TerminalNetworkSettings::TerminalNetworkSettings() { loadDefaults(); }

api_driver::obj::TerminalNetworkSettings::TerminalNetworkSettings(const TerminalNetworkSettings &src) = default;

api_driver::obj::TerminalNetworkSettings & api_driver::obj::TerminalNetworkSettings::operator=(const TerminalNetworkSettings &src) = default;

void api_driver::obj::TerminalNetworkSettings::loadDefaults() {
    managementIp = "0.0.0.0";
    managementGateway = "";
    isL2 = true;
    dataIp = "0.0.0.0";
    dataMtu = 1500;
    serverName = DEFAULT_SERVER_NAME;
}

void api_driver::obj::TerminalNetworkSettings::updateCallback(proxy::CpProxy &cp) {
    loadDefaults();
    try {
        managementIp = cp.getNetwork("addr");
        // s.managementIp += "/";
        // s.managementIp += std::to_string(calculateSubnetMask(cp.getNetwork("mask")));

        managementGateway = cp.getNetwork("gateway");
        if (cp.getNetwork("mode") == "tun") {
            isL2 = false;
            dataIp = cp.getNetwork("addr_data");
        } else {
            isL2 = true;
        }
        dataMtu = 1500;
        serverName = cp.getNetwork("name_serv");
        if (serverName.empty()) {
            serverName = DEFAULT_SERVER_NAME;
        }
    } catch (std::exception& e) {
        throw std::runtime_error(std::string("api_driver::obj::TerminalNetworkSettings::updateCallback() error: ") + e.what());
    }
}

void api_driver::obj::TerminalNetworkSettings::updateFromPt(boost::property_tree::ptree &pt) {
}

void api_driver::obj::TerminalNetworkSettings::store(proxy::CpProxy& cp) {
    try {
        cp.setNetwork("mode", isL2 ? "tap" : "tun");
        cp.setNetwork("addr", managementIp);

        if (!isL2) {
            cp.setNetwork("addr_data", dataIp);
        }
        cp.setNetwork("mask", "255.255.255.0");
        cp.setNetwork("gateway", managementGateway);

        // cp.setNetwork("data_mtu", std::to_string(dataMtu));

        cp.setNetwork("name_serv", serverName);
    } catch (std::exception& e) {
        throw std::runtime_error(std::string("api_driver::obj::TerminalNetworkSettings::store() error: ") + e.what());
    }
}

std::string api_driver::obj::TerminalNetworkSettings::asJson() {
    std::stringstream out;

    return out.str();
}

api_driver::obj::TerminalNetworkSettings::~TerminalNetworkSettings() = default;
#endif

#ifdef API_OBJECT_QOS_SETTINGS_ENABLE
api_driver::obj::TerminalQosSettings::TerminalQosSettings() = default;
api_driver::obj::TerminalQosSettings::TerminalQosSettings(const TerminalQosSettings &src) = default;
api_driver::obj::TerminalQosSettings & api_driver::obj::TerminalQosSettings::operator=(const TerminalQosSettings &src) = default;

void api_driver::obj::TerminalQosSettings::loadDefaults() {
    qosEnabled = false;
    qosSettingsJson = DEFAULT_QOS_CLASSES;
}

void api_driver::obj::TerminalQosSettings::updateCallback(proxy::CpProxy &cp) {
}

void api_driver::obj::TerminalQosSettings::updateFromPt(boost::property_tree::ptree &pt) {
}

void api_driver::obj::TerminalQosSettings::store(proxy::CpProxy &cp) {
}

std::string api_driver::obj::TerminalQosSettings::asJson() {
}

api_driver::obj::TerminalQosSettings::~TerminalQosSettings() = default;
#endif


api_driver::obj::TerminalState::TerminalState() = default;

void api_driver::obj::TerminalState::updateCallback(proxy::CpProxy& cp) {
    modulator_state mod{};
    modulator_settings modSet{};
    demodulator_state demod{};
#ifdef MODEM_IS_SCPC
    CinC_state cinc{};
#endif
    try {
        cp.getModState(mod);
        cp.getDemodState(demod);
        cp.getModSettings(modSet);

        fTxState = modSet.tx_is_on;
        fIsTest = modSet.tx_is_on && (!modSet.is_carrier || modSet.is_test_data);
#ifdef MODEM_IS_SCPC
        fIsCinC = modSet.is_cinc;
        if (fIsCinC) {
            cp.getCincState(cinc);
        }
#endif

#ifdef MODEM_IS_TDMA
        fInitState = cp.getDmaDebug("status_init");
#endif
    } catch (std::exception& e) {
        throw std::runtime_error(std::string("api_driver::obj::TerminalState::updateCallback() error: ") + e.what());
    }

    fRxState = demod.locks.sym_sync_lock && demod.locks.freq_lock && demod.locks.afc_lock && demod.locks.pkt_sync;
    fRxSymSyncLock = demod.locks.sym_sync_lock;
    fRxFreqSearchLock = demod.locks.freq_lock;
    fRxAfcLock = demod.locks.afc_lock;
    fRxPktSync = demod.locks.pkt_sync;

    fRxSnr = demod.snr;
    fRxRssi = demod.rssi;
    fRxModcod = demod.modcod;
    fRxFrameSizeNormal = !demod.is_short;
    fRxIsPilots = demod.is_pilots;

    fRxSymError = demod.sym_err;
    fRxFreqErr = demod.crs_freq_err;
    fRxFreqErrAcc = demod.fine_freq_err;
    fRxInputSignalLevel = demod.if_overload;
    fRxPllError = demod.afc_err;
    fRxSpeedOnRxKbit = static_cast<double>(demod.speed_in_bytes_rx) / 128.0;
    fRxSpeedOnIifKbit = static_cast<double>(demod.speed_in_bytes_rx_iface) / 128.0;
    fRxPacketsOk = demod.packet_ok_cnt;
    fRxPacketsBad = demod.packet_bad_cnt;
    fRxPacketsDummy = demod.dummy_cnt;

    fTxModcod = mod.modcod;
    fTxSpeedOnTxKbit = static_cast<double>(mod.speed_in_bytes_tx) / 128.0;
    fTxSpeedOnIifKbit = static_cast<double>(mod.speed_in_bytes_tx_iface) / 128.0;
#ifdef MODEM_IS_SCPC
    fTxSnr = mod.snr_remote;
    fTxFrameSizeNormal = !mod.is_short;
    fTxIsPilots = mod.is_pilots;
    fCincOcc = cinc.ratio_signal_signal;
    fCincCorrelator = cinc.carrier_lock;
    fCincCorrelatorFails = cinc.cnt_bad_lock;
    fCincFreqErr = cinc.freq_error_offset;
    fCincFreqErrAcc = cinc.freq_fine_estimate;
    fCincChannelDelay = cinc.delay_dpdi;
#endif
#ifdef MODEM_IS_TDMA
    fStatTxCenterFreq = modSet.central_freq_in_kGz;
    fStatTxSymSpeed = static_cast<double>(modSet.baudrate) / 1000.0;
#endif
}

nlohmann::json api_driver::obj::TerminalState::asJson() {
    nlohmann::json res{};

    res["initState"] = fInitState;
    res["testState"] = fIsTest;
    res["fTxState"] = fIsTest;
#ifdef MODEM_IS_SCPC
    res["isCinC"] = fIsCinC;
#endif

    res["rx"]["state"] = fRxState;
    res["rx"]["sym_sync_lock"] = fRxSymSyncLock;
    res["rx"]["freq_search_lock"] = fRxFreqSearchLock;
    res["rx"]["afc_lock"] = fRxAfcLock;
    res["rx"]["pkt_sync"] = fRxPktSync;

    res["rx"]["snr"] = fRxSnr;
    res["rx"]["rssi"] = fRxRssi;
    res["rx"]["modcod"] = fRxModcod;
    res["rx"]["frameSizeNormal"] = fRxFrameSizeNormal;
    res["rx"]["isPilots"] = fRxIsPilots;

    res["rx"]["symError"] = fRxSymError;
    res["rx"]["freqErr"] = fRxFreqErr;
    res["rx"]["freqErrAcc"] = fRxFreqErrAcc;
    res["rx"]["inputSignalLevel"] = fRxInputSignalLevel;
    res["rx"]["pllError"] = fRxPllError;
    res["rx"]["speedOnRxKbit"] = fRxSpeedOnRxKbit;
    res["rx"]["speedOnIifKbit"] = fRxSpeedOnIifKbit;
    res["rx"]["packetsOk"] = fRxPacketsOk;
    res["rx"]["packetsBad"] = fRxPacketsBad;
    res["rx"]["packetsDummy"] = fRxPacketsDummy;

    res["tx"]["modcod"] = fTxModcod;
    res["tx"]["speedOnTxKbit"] = fTxSpeedOnTxKbit;
    res["tx"]["speedOnIifKbit"] = fTxSpeedOnIifKbit;
#ifdef MODEM_IS_SCPC
    res["tx"]["snr"] = fTxSnr;
    res["tx"]["frameSizeNormal"] = fTxFrameSizeNormal;
    res["tx"]["isPilots"] = fTxIsPilots;

    if (fIsCinC) {
        if (fTxState) {
            res["cinc"]["correlator"] = fCincCorrelator;
        } else {
            res["cinc"]["correlator"] = nullptr;
        }

        res["cinc"]["occ"] = fCincOcc;
        res["cinc"]["correlatorFails"] = fCincCorrelatorFails;
        res["cinc"]["freqErr"] = fCincFreqErr;
        res["cinc"]["freqErrAcc"] = fCincFreqErrAcc;
        res["cinc"]["channelDelay"] = fCincChannelDelay;
    } else {
        res["cinc"]["correlator"] = nullptr;
    }
#endif
#ifdef MODEM_IS_TDMA
    res["tx"]["centerFreq"] = fTxCenterFreq;
    res["tx"]["symSpeed"] = fTxSymSpeed;
#endif

    return res;
}

api_driver::obj::TerminalState::~TerminalState() = default;


api_driver::obj::TerminalDeviceState::TerminalDeviceState() = default;
api_driver::obj::TerminalDeviceState::TerminalDeviceState(const TerminalDeviceState &src) = default;
api_driver::obj::TerminalDeviceState & api_driver::obj::TerminalDeviceState::operator=(const TerminalDeviceState &src) = default;

void api_driver::obj::TerminalDeviceState::updateCallback(proxy::CpProxy &cp) {
    {
        device_state ds{};
        cp.getDeviceState(ds);
        fTempAdrv = ds.adrv_temp;
        fTempZynq = ds.pl_temp;
        fTempFpga = ds.zynq_temp;
    }
#ifdef MODEM_IS_TDMA
    {
        progress_msg ds{};
        logCpApiError("api_driver::TerminalApiDaemon::updateState()->CP_GetUpdateStatus()", CP_GetUpdateStatus(sid, ds));
        fOtaStatus = ds.status;
        fOtaPercent = ds.dwl_percent;
        fOtaImage = ds.cur_image;
    }
#endif

    struct sysinfo info{};
    sysinfo(&info);

    const double f_load = 100.0 / ((1 << SI_LOAD_SHIFT) * get_nprocs());
    fOsUptime = info.uptime;
    fOsLoad1 = f_load * static_cast<double>(info.loads[0]);
    fOsLoad5 = f_load * static_cast<double>(info.loads[1]);
    fOsLoad15 = f_load * static_cast<double>(info.loads[2]);
    fOsTotalram = (info.totalram * info.mem_unit) >> 20; // Mb
    fOsFreeram = (info.totalram * info.mem_unit) >> 20; // Mb
    fOsProcs = info.procs;
}

nlohmann::json api_driver::obj::TerminalDeviceState::asJson() const {
    nlohmann::json res;
    res["uptime"] = fOsUptime;
    res["load1min"] = fOsLoad1;
    res["load5min"] = fOsLoad5;
    res["load15min"] = fOsLoad15;
    res["totalram"] = fOsTotalram;
    res["freeram"] = fOsFreeram;
    res["procs"] = fOsProcs;
    res["adrv"] = fTempAdrv;
    res["fpga"] = fTempFpga;
    res["zynq"] = fTempZynq;
#ifdef MODEM_IS_TDMA
    if (fUpgradeImage.empty()) {
        res["upgradeStatus"] = "Нет обновлений";
        res["upgradePercent"] = 0;
        res["upgradeImage"] = "";
    } else {
        switch (fUpgradeStatus) {
            case NORM_RX_OBJECT_NEW_API: res["upgradeStatus"] = "Начало загрузки"; break;
            case NORM_RX_OBJECT_INFO_API: res["upgradeStatus"] = "Получено имя образа"; break;
            case NORM_RX_OBJECT_UPDATED_API: res["upgradeStatus"] = "Загружается"; break;
            case NORM_RX_OBJECT_COMPLETED_API: res["upgradeStatus"] = "Загрузка завершена"; break;
            case NORM_RX_OBJECT_ABORTED_API: res["upgradeStatus"] = "Загрузка прервана"; break;
            default: res["upgradeStatus"] = "?";
        }
        res["upgradePercent"] = fUpgradePercent;
        res["upgradeImage"] = fUpgradeImage;
    }

#endif
    return res;
}

api_driver::obj::TerminalDeviceState::~TerminalDeviceState() = default;





